/******************************************************************************
 * sdla_ec.h	
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * Copyright:	(c) 1995-2001 Sangoma Technologies Inc.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 * ============================================================================
 ******************************************************************************
 */

#ifndef __SDLA_EC_H
# define __SDLA_EC_H

#define AFT_HWEC_CLEAR_RESET	1
#define AFT_HWEC_SET_RESET	2

#define WAN_OCT6100_RC_OK				0x0000
#define WAN_OCT6100_RC_CPU_INTERFACE_NO_RESPONSE	0x0001
/* OCT6100 state machine */
enum {
	WAN_OCT6100_STATE_NONE = 0x00,
	WAN_OCT6100_STATE_RESET,
	WAN_OCT6100_STATE_READY,
	WAN_OCT6100_STATE_CHIP_READY
};
#define WAN_OCT6100_STATE_DECODE(state)				\
	(state == WAN_OCT6100_STATE_RESET) ? "Reset" :		\
	(state == WAN_OCT6100_STATE_READY) ? "Ready" :		\
	(state == WAN_OCT6100_STATE_CHIP_READY) ? "Chip Ready" :\
					"Unknown"

/* OCT6100 api command */
#if 1
# define WAN_OCT6100_CMD_NONE			_IOWR('E', 1, struct wan_ec_api_)
# define WAN_OCT6100_CMD_GET_INFO		_IOWR('E', 2, struct wan_ec_api_)
# define WAN_OCT6100_CMD_SET_INFO		_IOWR('E', 3, struct wan_ec_api_)
# define WAN_OCT6100_CMD_CLEAR_RESET		_IOWR('E', 4, struct wan_ec_api_)
# define WAN_OCT6100_CMD_SET_RESET		_IOWR('E', 5, struct wan_ec_api_)
# define WAN_OCT6100_CMD_BYPASS_ENABLE		_IOWR('E', 6, struct wan_ec_api_)
# define WAN_OCT6100_CMD_BYPASS_DISABLE		_IOWR('E', 7, struct wan_ec_api_)
# define WAN_OCT6100_CMD_API_WRITE		_IOWR('E', 8, struct wan_ec_api_)
# define WAN_OCT6100_CMD_API_WRITE_SMEAR	_IOWR('E', 9, struct wan_ec_api_)
# define WAN_OCT6100_CMD_API_WRITE_BURST	_IOWR('E', 10, struct wan_ec_api_)
# define WAN_OCT6100_CMD_API_READ		_IOWR('E', 11, struct wan_ec_api_)
# define WAN_OCT6100_CMD_API_READ_BURST		_IOWR('E', 12, struct wan_ec_api_)
#else
enum {
	WAN_OCT6100_CMD_NONE		= 0x00,
	WAN_OCT6100_CMD_GET_INFO,
	WAN_OCT6100_CMD_SET_INFO,
	WAN_OCT6100_CMD_CLEAR_RESET,
	WAN_OCT6100_CMD_SET_RESET,
	WAN_OCT6100_CMD_BYPASS_ENABLE,
	WAN_OCT6100_CMD_BYPASS_DISABLE,
	WAN_OCT6100_CMD_API_WRITE,
	WAN_OCT6100_CMD_API_WRITE_SMEAR,
	WAN_OCT6100_CMD_API_WRITE_BURST,
	WAN_OCT6100_CMD_API_READ,
	WAN_OCT6100_CMD_API_READ_BURST
};
#endif

#define WAN_OCT6100_CMD_DECODE(cmd)							\
	(cmd == WAN_OCT6100_CMD_GET_INFO) ?		"Get Info" :			\
	(cmd == WAN_OCT6100_CMD_SET_INFO) ?		"Set Info" :			\
	(cmd == WAN_OCT6100_CMD_CLEAR_RESET) ?		"Clear Reset" :			\
	(cmd == WAN_OCT6100_CMD_SET_RESET) ?		"Set Reset" :			\
	(cmd == WAN_OCT6100_CMD_BYPASS_ENABLE) ?	"Bypass Enable" :		\
	(cmd == WAN_OCT6100_CMD_BYPASS_DISABLE) ?	"Bypass Disable" :		\
	(cmd == WAN_OCT6100_CMD_API_WRITE) ?		"Oct6100 API WRITE" :		\
	(cmd == WAN_OCT6100_CMD_API_WRITE_SMEAR) ?	"Oct6100 API WRITE SMEAR" :	\
	(cmd == WAN_OCT6100_CMD_API_WRITE_BURST) ?	"Oct6100 API WRITE BURST" :	\
	(cmd == WAN_OCT6100_CMD_API_READ) ?		"Oct6100 API READ" :		\
	(cmd == WAN_OCT6100_CMD_API_READ_BURST) ?	"Oct6100 API READ BURST" :	\
							"Unknown"

#if 0
enum {
	WAN_OCT6100_BYPASS	= 0x01,
	WAN_OCT6100_BYPASS_FULL,
	WAN_OCT6100_BYPASS_LINE,
};
#endif

/*===========================================================================*\
  User process context - This would probably have to be defined elsewhere.
  This structure must be allocated by the calling process and the parameters
  should be defined prior to open the OCT6100 chip.
\*===========================================================================*/
typedef struct _OCTPCIDRV_USER_PROCESS_CONTEXT_
{
	/* Is main process */
	unsigned int	fMainProcess;

	/* Handle to driver (opened by calling process) */
	unsigned int	hDriver;

	/* Interface name to driver (copied by calling process) */
	unsigned char	devname[WAN_DRVNAME_SZ+1];
	unsigned char	ifname[WAN_IFNAME_SZ+1];

	/* Handle to serialization object used for read and writes */
	unsigned int 	ulUserReadWriteSerObj;

	/* Board index. */
	unsigned int	ulBoardId;

	/* Board type. */
	unsigned int	ulBoardType;

} tOCTPCIDRV_USER_PROCESS_CONTEXT, *tPOCTPCIDRV_USER_PROCESS_CONTEXT;

#define u_ecd			u.ecd
#define u_oct6100_write		u.oct6100_write
#define u_oct6100_write_swear	u.oct6100_write_swear
#define u_oct6100_write_burst	u.oct6100_write_burst
#define u_oct6100_read		u.oct6100_read
#define u_oct6100_read_burst	u.oct6100_read_burst
typedef struct wan_ec_api_ {
	char		name[WAN_DRVNAME_SZ+1];
	char		devname[WAN_DRVNAME_SZ+1];
	char		if_name[WAN_IFNAME_SZ+1];
	unsigned int	cmd;
	unsigned int	type;
	unsigned int	ulResult;
	unsigned int	dummy;
	
	union {
		struct ecd_ {
			unsigned int	state;
			unsigned char	fe_media;
			unsigned int	fe_lineno;
			unsigned int	fe_max_channels;
			unsigned int	fe_tdmv_law;
			unsigned int	channel;
			unsigned int	chip_no;
			unsigned int	max_channels;
			unsigned int	max_ec_channels;
			unsigned int	ulApiInstanceSize;
		} ecd;
		struct oct6100_write_ {
			unsigned int	ulChipIndex;
			unsigned int	ulWriteAddress;
			unsigned short	usWriteData;
		} oct6100_write;
		struct oct6100_write_swear_ {
			unsigned int	ulChipIndex;
			unsigned int	ulWriteAddress;
			unsigned short	usWriteData;
			unsigned int	ulWriteLength;
		} oct6100_write_swear;
		struct oct6100_write_burst_ {
			unsigned int	ulChipIndex;
			unsigned int	ulWriteAddress;
			unsigned int	ulWriteLength;
			unsigned int	pWriteData;	//unsigned short *pWriteData;
		} oct6100_write_burst;
		struct oct6100_read_ {
			unsigned int	ulChipIndex;
			unsigned int	ulReadAddress;
			unsigned short	usReadData;
		} oct6100_read;
		struct oct6100_read_burst_ {
			unsigned int	ulChipIndex;
			unsigned int	ulReadAddress;
			unsigned int	ulReadLength;
			unsigned int	pReadData; 	//unsigned short *pReadData
		} oct6100_read_burst;
	} u;
} wan_ec_api_t;

typedef struct wan_ec_iface_
{
	int	(*action)(void *pcard, int, int);

} wan_ec_iface_t;

#if defined(WAN_KERNEL)
/* global interface functions */
void *wan_ec_config (void *pcard, int max_channels);
int wan_ec_remove(void*, void *pcard);
int wan_ec_ioctl(void*, struct ifreq*, void *pcard);
int wan_ec_dev_ioctl(void*, void *pcard);

int wanpipe_ecdev_reg(void *ec, char *ec_name, int ec_chip_no);
int wanpipe_ecdev_unreg(int ec_chip_no);

#endif

#endif /* __SDLA_EC_H */
