/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdlamain.c,v 1.27 2006/03/06 15:03:45 sangoma Exp $
 */

/*
 ****************************************************************************
 * sdlamain.c	WANPIPE(tm) Multiprotocol WAN Link Driver.  Main module.
 *
 * Author:	Alex Feldman    <al.feldman@sangoma.com> 
 *
 * ===========================================================================
 * Jul 31, 2000 Alex Feldman    The OpenBSD-2.6/2.7 kernel has bug during 
 *				detach network interface if NAT is enabled.
 *				Do not detach network interface from the 
 *				kernel in 2.6/2.7 until fixing NAX bug.
 *				(label NAT_OBSD_KERNEL_BUG)
 * Nov 20, 1999	Alex Feldman 	Updated to FreeBSD 3.2
 * Jan 02, 1997	Gene Kozin	Initial version.
 ****************************************************************************
*/
/*
 ****************************************************************************
 ****		For Debug purpose (only OpenBSD)			 ****
 ****************************************************************************
*/
#ifdef __OpenBSD__
# undef WANDEBUG		/* Uncomment this line for Debug purpose */
#endif

#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/wanpipe_defines.h>
#include <net/wanpipe_debug.h>
#include <net/wanpipe.h>	/* WANPIPE common user API definitions */
#include <net/sdlapci.h>

/****** Defines & Macros ****************************************************/

#define	MAX_CMD_RETRY	10		/* max number of firmware retries */

#ifndef CONFIG_PRODUCT_WANPIPE_FR
  #define wpf_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_CHDLC
  #define wpc_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_X25
 #define wpx_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_PPP
 #define wpp_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_MULTPROT 
 #define wsppp_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_MULTFR
 #define wp_hdlc_fr_init(a,b) (-EPROTONOSUPPORT)	
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_ADSL
 #define wp_adsl_init(card,conf) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_SDLC
 #define wp_sdlc_init(a,b) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_EDU
 #define wpedu_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_BSC
 #define wpbsc_init(a,b) (-EPROTONOSUPPORT)
#endif
 
#ifndef CONFIG_PRODUCT_WANPIPE_SS7
  #define wpss7_init(a,b) (-EPROTONOSUPPORT) 
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_BSCSTRM
 #define wp_bscstrm_init(a,b) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_BITSTRM
 #define wpbit_init(a,b) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_MLINK_PPP
 #define wph_init(a,b) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_AFT
 #define wp_xilinx_init(card,conf) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_AFT_TE1
 #define wp_aft_te1_init(card,conf) (-EPROTONOSUPPORT)
#endif

#ifndef CONFIG_PRODUCT_WANPIPE_AFT_TE3
 #define wp_aft_te3_init(card,conf) (-EPROTONOSUPPORT)
#endif


/****** Function Prototypes *************************************************/

/* Loadable kernel module function interface  */
int wanpipe_init(void *arg);
int wanpipe_exit(void *arg);
#if defined(__FreeBSD__)
WAN_MODULE_DEFINE(
		wanpipe, "wanpipe",
		"Alex Feldman <al.feldman@sangoma.com>",
		"WANPIPE(tm) Multiprotocol Driver",
		"GPL",
		wanpipe_init, wanpipe_exit, 
		NULL);
WAN_MODULE_DEPEND(wanpipe,wanrouter,1,WANROUTER_MAJOR_VER,WANROUTER_MAJOR_VER);
WAN_MODULE_DEPEND(wanpipe,sdladrv,1,SDLADRV_MAJOR_VER,SDLADRV_MAJOR_VER);
# if defined(CONFIG_PRODUCT_WANPIPE_TDM_VOICE)
WAN_MODULE_DEPEND(wanpipe,zaptel,1,1,1);
# endif
#else
extern int wanrouter_init(void*);
extern int wanrouter_exit(void*);
extern int sdladrv_init(void*);
extern int sdladrv_exit(void*);
extern int wanpipe_lip_init(void*);
extern int wanpipe_lip_exit(void*);
extern struct cdevsw wanrouter_devsw;
WAN_MODULE_DEFINE(
		wanpipe, "wanpipe",
		"Alex Feldman <al.feldman@sangoma.com>",
		"WANPIPE(tm) Multiprotocol Driver",
		"GPL",
		wanpipe_init, wanpipe_exit, 
		&wanrouter_devsw);
#endif

WAN_MODULE_VERSION(wanpipe, WANPIPE_MAJOR_VER);

#if 0
------------------
#if defined(__FreeBSD__)
static int wanpipe_load __P((module_t, int, void*));
/* Now declare the module to the system */
DEV_MODULE(wanpipe, wanpipe_load, NULL);
#if 0
// FreeBSD-3.x/4.x
static moduledata_t wanpipe_mod = {
	"wanpipe",
	wanpipe_load,
	0
};
DECLARE_MODULE(wanpipe, wanpipe_mod, SI_SUB_KLD, SI_ORDER_MIDDLE);
#endif
MODULE_DEPEND(wanpipe, wanrouter, 1, WANROUTER_MAJOR_VER,WANROUTER_MAJOR_VER);
MODULE_DEPEND(wanpipe, sdladrv, 1, SDLADRV_MAJOR_VER,SDLADRV_MAJOR_VER);
MODULE_VERSION(wanpipe, WANPIPE_MAJOR_VER);
#elif defined(__OpenBSD__)
extern int wanrouter_init(void);
extern int wanrouter_exit(void);
extern int sdladrv_init(void);
extern int sdladrv_exit(void);
extern int wanpipe_lip_init(void);
extern int wanpipe_lip_exit(void);
extern struct cdevsw wanrouter_devsw;
int wanpipe_load(struct lkm_table* lkmtp, int cmd);
MOD_DEV("wanpipe", LM_DT_CHAR, -1, &wanrouter_devsw);
#elif defined(__NetBSD__)
# if (__NetBSD_Version__ < 200000000)
extern int wanrouter_init(void);
extern int wanrouter_exit(void);
extern int sdladrv_init(void);
extern int sdladrv_exit(void);
extern int wanpipe_lip_init(void);
extern int wanpipe_lip_exit(void);
extern struct cdevsw wanrouter_devsw;
int wanpipe_load(struct lkm_table* lkmtp, int cmd);
MOD_DEV("wanpipe", LM_DT_CHAR, -1, &wanrouter_devsw);
# else
extern int wanrouter_init(void);
extern int wanrouter_exit(void);
extern int sdladrv_init(void);
extern int sdladrv_exit(void);
extern int wanpipe_lip_init(void);
extern int wanpipe_lip_exit(void);
extern struct cdevsw wanrouter_devsw;
int wanpipe_load(struct lkm_table* lkmtp, int cmd);
MOD_DEV("wanpipe", "wanpipe", NULL, -1, &wanrouter_devsw, -1);
# endif
#endif
---------------------
#endif

/* WAN link driver entry points */
static int setup    (wan_device_t* wandev, wandev_conf_t* conf);
static int shutdown (wan_device_t* wandev, wandev_conf_t* conf);
static int ioctl    (wan_device_t* wandev, u_long cmd, caddr_t arg);
static int debugging (wan_device_t* wandev);

/* IOCTL handlers */
static int ioctl_dump	(sdla_t* card, sdla_dump_t* u_dump);
static int ioctl_exec	(sdla_t* card, sdla_exec_t* u_exec);

/* ISA/PCI interrupt handler */
static void 	sdla_isr(void *arg);

/* Miscellaneous functions */
static void release_hw(sdla_t *card);

/* Copy packets to mbuf chains */
struct mbuf* sdla_get(struct ifnet* ifp, char* data, unsigned totlen);
int sdla_reply_udp(unsigned char *data, unsigned int mbox_len);

/* SDLA ISA/PCI varibles */
extern int number_cards;		/* total number of SDLA cards */
extern unsigned int number_S514_cards;  /* total number of S514 cards */

static int check_s508_conflicts (sdla_t* card,wandev_conf_t* conf, int*);
static int check_s514_conflicts (sdla_t* card,wandev_conf_t* conf, int*);
static int check_adsl_conflicts (sdla_t* card,wandev_conf_t* conf, int*);
static int check_aft_conflicts(sdla_t* card,wandev_conf_t* conf, int *irq);
/****** Global Data **********************************************************
 * Note: All data must be explicitly initialized!!!
 */
#ifdef __OpenBSD__
extern int ticks;
#endif

/* private data */
#define WAN_DRV_NAME	"wanpipe"

char*	wan_drvname = "wanpipe"; 
static char wan_fullname[]	= "WANPIPE(tm) Multiprotocol Driver";
static char wan_copyright[]	= "(c) 1995-2001 Sangoma Technologies Inc.";

static int Sangoma_cards_no = 0;
static int active = 0;			/* number of active cards */
static sdla_t* card_array = NULL;	/* adapter data space */
sdla_t*	wanpipe_debug = NULL;

/******* Kernel Loadable Module Entry Points ********************************/
/* 
** ============================================================================
** Module 'insert' entry point.
** o print announcement
** o allocate adapter data space
** o initialize static data
** o calibrate SDLA shared memory access delay.
*/
int wanpipe_init (void *arg)
{
	int cnt = 0, err = 0;

#if defined(__NetBSD__) || defined(__OpenBSD__)
	sdladrv_init(arg);
	wanrouter_init(arg);
	wanpipe_lip_init(arg);
#endif
        DEBUG_EVENT("%s v%s %s\n", 
				wan_fullname, 
				WANPIPE_VERSION, 
				wan_copyright); 

	/* Verify number of cards and allocate adapter data space */
	Sangoma_cards_no = sdla_hw_probe();
	if (Sangoma_cards_no){
		DEBUG_EVENT("WANPIPE: Allocating maximum %d devices:\n",
				Sangoma_cards_no);
		DEBUG_EVENT("WANPIPE: wanpipe1-wanpipe%d.\n",
				Sangoma_cards_no);
	}else{
		DEBUG_EVENT("WANPIPE: No Sangoma cards found, unloading modules!\n");
		return -ENODEV;		
	}
	card_array = wan_malloc(sizeof(sdla_t)*Sangoma_cards_no);
	if (card_array == NULL){
		return -ENOMEM;
	}
	bzero(card_array, sizeof(sdla_t)*Sangoma_cards_no); 
	/* Register adapters with WAN router */
	for (cnt = 0; cnt < Sangoma_cards_no; ++ cnt){ 
		sdla_t* card  	   = &card_array[cnt];
		wan_device_t* wandev   = &card->wandev;

		sprintf(card->devname, "%s%d", WAN_DRV_NAME, cnt+1);
		card->next 	  	   = NULL;
		wandev->magic          = ROUTER_MAGIC;
		wandev->name           = card->devname;
		wandev->private        = card;
		wandev->enable_tx_int  = 0;
		wandev->setup          = &setup;
		wandev->shutdown       = &shutdown;
		wandev->ioctl          = &ioctl;
		wandev->debugging      = &debugging;
		wandev->debug_read      = &wan_debug_read;
		err = register_wan_device(wandev);
		if (err){
			DEBUG_EVENT("%s: Registration failed with error %d!\n",
					 card->devname, err);
			break;
		}
	}
	if (cnt){
		Sangoma_cards_no = cnt;	/* adjust actual number of cards */
	}else{
		wan_free(card_array);
		DEBUG_EVENT("*** Init Module: No cards registered.\n"); 
		return -ENODEV;
	}
	
	/* Initialize global data */
	wanpipe_globals_util_init();

	aft_global_hw_device_init();

	return err;
}

/* 
** ============================================================================
** Module 'remove' entry point.
** o release all remaining system resources
*/
 int wanpipe_exit(void *arg)
{
	int 	i;

	if (card_array != NULL){
		for(i = 0; i < Sangoma_cards_no; i++){
			sdla_t *card = &card_array[i];
			unregister_wan_device(card->devname);
		}
		wan_free(card_array);
	}
#if defined(__NetBSD__) || defined(__OpenBSD__)
	wanpipe_lip_exit(arg);
	wanrouter_exit(arg);
	sdladrv_exit(arg);
#endif
	return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/* 
 * ============================================================================
 * Setup/configure WAN link driver.
 * o check adapter state
 * o make sure firmware is present in configuration
 * o make sure I/O port and IRQ are specified
 * o make sure I/O region is available
 * o allocate interrupt vector
 * o setup SDLA hardware
 * o call appropriate routine to perform protocol-specific initialization
 * o mark I/O region as used
 * o if this is the first active card, then schedule background task
 *
 * This function is called when router handles ROUTER_SETUP IOCTL. The
 * configuration structure is in kernel memory (including extended data, if
 * any).
 */
 
static int setup (wan_device_t* wandev, wandev_conf_t* conf)
{
	sdla_t*	card = NULL;
	int 	err = 0, irq = 0;

	WAN_ASSERT(wandev == NULL);
	WAN_ASSERT(wandev->private == NULL);
	WAN_ASSERT(conf == NULL);
	
	card = wandev->private;
	if (wandev->state != WAN_UNCONFIGURED){
		DEBUG_EVENT("%s: Device already configured!\n", 
						card->devname);
		return -EBUSY;		/* already configured */
	}

	DEBUG_EVENT("\nProcessing WAN device %s...\n", card->devname);

	/* Initialize counters for each device.
	 * Used for counting number if times new interface and delete interface
	 * called.
	 */
	wandev->new_if_cnt = 0;
	wandev->del_if_cnt = 0;
	wandev->config_id = conf->config_id;
	switch(conf->config_id){
	case WANCONFIG_AFT:
		conf->card_type = WANOPT_AFT;
		break;
	case WANCONFIG_AFT_TE1:
		conf->card_type = WANOPT_AFT104;
		conf->S514_CPU_no[0] = 'A';
		break;
	case WANCONFIG_AFT_ANALOG:
		conf->card_type = WANOPT_AFT_ANALOG;
		conf->S514_CPU_no[0] = 'A';
		break;
	case WANCONFIG_AFT_TE3:
		conf->card_type = WANOPT_AFT300;
		conf->S514_CPU_no[0] = 'A';
		break;
	}
	wandev->card_type = conf->card_type;
	card->hw = sdla_register(&card->hw_iface, conf, card->devname);
	if (card->hw == NULL){
		DEBUG_EVENT("%s: ERROR, invalid card type!\n",card->devname);
		return -EINVAL;
	}

	/* Check for resource conflicts and setup the
	** card for piggibacking if necessary 
	*/
	switch (conf->card_type){
	case WANOPT_S50X:
		if (!conf->data_size || (conf->data == NULL)){
			DEBUG_EVENT("%s: firmware not found in configuration data!\n",
						card->devname);
			err = -EINVAL;
			goto setup_failed;
		}

		if ((err=check_s508_conflicts(card,conf,&irq)) != 0){
			goto setup_failed;
		}
		card->type = SDLA_S508;
		break;
			
	case WANOPT_S51X:
		if ((!conf->data_size || (conf->data == NULL)) && (conf->config_id != WANCONFIG_DEBUG)){
			DEBUG_EVENT("%s: firmware not found in configuration data!\n",
						card->devname);
			err = -EINVAL;
			goto setup_failed;
		}

		if ((err=check_s514_conflicts(card,conf,&irq)) != 0){
			goto setup_failed;
		}
		card->type		    = SDLA_S514;
		card->wandev.S514_cpu_no[0] = conf->S514_CPU_no[0];
		break;
		
	case WANOPT_ADSL:
		if ((err=check_adsl_conflicts(card,conf,&irq)) != 0){
			goto setup_failed;
		}
		card->type = SDLA_ADSL;
		break;

	case WANOPT_AFT:
	case WANOPT_AFT104:
	case WANOPT_AFT300:
	case WANOPT_AFT_ANALOG:
		if ((err=check_aft_conflicts(card,conf,&irq)) != 0){
			goto setup_failed;
		}
		card->type = SDLA_AFT;
		break;

	default:
		DEBUG_EVENT("%s: ERROR, invalid card type!\n",
					card->devname);
		err = -EINVAL;
		goto setup_failed;
	}
 
	card->hw_iface.getcfg(card->hw, SDLA_CARDTYPE, &card->type);
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &card->adptr_type);
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERSUBTYPE, &card->adptr_subtype);

	if (!card->wandev.piggyback && !card->configured){
		/* Configure hardware, load firmware, etc. */
		DEBUG_EVENT("%s: Starting hardware setup...\n", 
					card->devname);
		err = card->hw_iface.setup(card->hw, conf);
		if (err){
			goto setup_failed;
		}
		if (conf->config_id == WANCONFIG_DEBUG){ 
			wanpipe_debug = card;
			wanpipe_debug->u.debug.total_len = 0;
			wanpipe_debug->u.debug.current_offset = 
					sizeof(wanpipe_kernel_msg_info_t);
			card->configured = 1;
			/* Forse it to be CONNECTED */
			card->wandev.state = WAN_CONNECTED;
			wan_spin_lock_init(&wanpipe_debug->wandev.lock);
			return 0;
		}
		DEBUG_EVENT("%s: Set interrupt handler...\n", 
					card->devname);
 		err = card->hw_iface.set_intrhand(
					card->hw,
					sdla_isr,
					(void*)card,
					FE_LINENO(&conf->fe_cfg));
		if (err){
			DEBUG_EVENT("%s: Failed set interrupt handler!\n", 
						card->devname);
			card->hw_iface.down(card->hw);
			sdla_unregister(&card->hw, card->devname);
			return -EINVAL;
			goto setup_failed; 
		}
	
        	/* Intialize WAN device data space */
	        wandev->dma     = 0;
	        wandev->irq	= conf->irq;

		wan_spin_lock_init(&wandev->lock);
	}else{
		if (card->wandev.piggyback == WANOPT_YES){
   			DEBUG_EVENT("%s: Card piggybacking (%d) to %s!\n",
					card->devname, 
					card->wandev.piggyback, 
					(card->next) ? 	card->next->devname : 
							"NULL");
		}else{
   			DEBUG_EVENT("%s: Card configured %ld!\n",
					card->devname, card->configured); 
		}
	}
	/* Hardware configuration is done */
	card->configured = 1;

	/* Read PCI adapter type */
	/*card->adapter_type = sdla_get_adapter_type(&card->hw);*/

	/* Protocol-specific initialization */
	switch (conf->config_id){ 
	case WANCONFIG_X25:
		err = wpx_init(card, conf);
		break;

	case WANCONFIG_FR:
		/* 
		 * Return priority to previous value, because in this function I run
		 * interrupt test.
		 */
		err = wpf_init(card, conf);
		break;

	case WANCONFIG_PPP:
		err = wpp_init(card, conf);
		break;

	case WANCONFIG_CHDLC:
		if (conf->ft1){
		    DEBUG_EVENT("%s: Starting FT1 CSU/DSU Config Driver.\n", 
					card->devname);
		    err = wpft1_init(card, conf);
		}else{
		    DEBUG_EVENT("%s: Starting CHDLC Protocol Init.\n", 
					card->devname);
		    err = wpc_init(card, conf);
		}
		break;

	case WANCONFIG_BSC:
		err = wpbsc_init(card, conf);
		break;

	case WANCONFIG_ADSL:
		DEBUG_EVENT("%s: Starting ADSL device.\n", 
					card->devname);
		err = wp_adsl_init(card,conf);
		break;
		
	case WANCONFIG_MPPP:
			
		DEBUG_EVENT("%s: Starting Multi-Port Protocol Driver Init.\n",
					card->devname);
		err = wpc_init(card,conf);
		break;
		
	case WANCONFIG_MLINK_PPP:
			
		DEBUG_EVENT("%s: Starting Multi-Link PPP Driver Init.\n",
					card->devname);
		err = wph_init(card,conf);
		break;
		
	case WANCONFIG_AFT:
		DEBUG_EVENT("%s: Starting AFT Hardware Init.\n",
					card->devname);
		err = wp_xilinx_init(card,conf);
		break;

	case WANCONFIG_AFT_TE1:
		DEBUG_EVENT("%s: Starting AFT Quad Hardware Init.\n",
					card->devname);
		err = wp_aft_te1_init(card,conf);
		break;

	case WANCONFIG_AFT_ANALOG:
		DEBUG_EVENT("%s: Starting AFT Analog Hardware Init.\n",
					card->devname);
		err = wp_aft_analog_init(card,conf);
		break;

	case WANCONFIG_AFT_TE3:
		DEBUG_EVENT("%s: Starting AFT TE3 Hardware Init.\n",
					card->devname);
		err = wp_aft_te3_init(card,conf);
		break;

	default:
	        DEBUG_EVENT("%s: this firmware is not supported!\n", 
					card->devname);
		err = -EINVAL;
	}

	if (err){ 
		if (err == -EPROTONOSUPPORT){
			DEBUG_EVENT("%s: Error, Protocol selected has not been compiled!\n",
					card->devname);
			DEBUG_EVENT("%s:      Re-configure the module Makefile and re-build the modules!\n",
					card->devname);
		}
		goto setup_failed;
	}

	WAN_DEBUG_INIT(card);
	
	if (conf->config_id == WANCONFIG_MLINK_PPP){
#if defined(NETGRAPH)
		err = wan_ng_init(card);
#else
		err = -EINVAL;
#endif
		if (err < 0){
			err = -EINVAL;
			goto setup_failed;
		}
	}

	active++;  /* increment number of active cards */

	DEBUG_CFG("%s: Device setup is finished\n", card->devname);
	return 0;

setup_failed:
	release_hw(card);
	sdla_unregister(&card->hw, card->devname);
	return err;
}

/*================================================================== 
 * configure_s508_card
 * 
 * For a S508 adapter, check for a possible configuration error in that
 * we are loading an adapter in the same IO port as a previously loaded S508
 * card.
 */ 

static int check_s508_conflicts (sdla_t* card,wandev_conf_t* conf, int *irq)
{
	sdla_t*		nxt_card = NULL;
	wan_spinlock_t	s;
	int 		i = 0;
	
	if (conf->ioport <= 0) {
		DEBUG_EVENT("%s: can't configure without I/O port address!\n",
						card->devname);
		return -EINVAL;
	}

	if (conf->irq <= 0) {
		DEBUG_EVENT("%s: can't configure without IRQ!\n",
						card->devname);
		return -EINVAL;
	}

	if (card->configured)
		return 0;

	/* 
	 * Check for already loaded card with the same IO port and IRQ.
   	 * If found, copy its hardware configuration and use its resources
	 * (i.e. piggybacking)
	 */
	for(i = 0; i < Sangoma_cards_no; i++){
	        nxt_card = &card_array[i];
		if (nxt_card == card){
		    continue;
		}

		if (nxt_card->hw == card->hw){
			if ((conf->config_id == WANCONFIG_CHDLC || conf->config_id == WANCONFIG_MPPP) && 
			    (nxt_card->wandev.config_id == WANCONFIG_CHDLC || nxt_card->wandev.config_id == WANCONFIG_MPPP)){

				if (nxt_card->next != NULL){
					sdla_t*	card1 = (sdla_t*)nxt_card->next;
					/* Trying to run piggibacking with a wrong protocol */
					DEBUG_EVENT("%s: ERROR: Resource busy: CPU %c PCISLOT %d\n",
							card->devname, 
							conf->S514_CPU_no[0],
							conf->PCI_slot_no);
					DEBUG_EVENT("%s:        These resources are already is used\n",
							card->devname);
					DEBUG_EVENT("%s:        by %s and %s devices!\n",
							card->devname, nxt_card->devname, card1->devname);
					return -EBUSY;
				}
				wan_spin_lock_irq(NULL, &s);
		    		/* ALEX *irq = nxt_card->hw.irq; */	
		    		nxt_card->next		= card;
		    		card->next		= nxt_card;
		    		card->wandev.piggyback	= WANOPT_YES;
				wan_spin_lock_init(&card->wandev.lock);
				wan_spin_unlock_irq(NULL, &s);
				break;
			}else{
				u16	ioport;
				/* Trying to run piggibacking with a wrong protocol */
				card->hw_iface.getcfg(card->hw, SDLA_IOPORT, &ioport);
				DEBUG_EVENT("%s: ERROR: Resource busy, ioport: 0x%x\n"
						 "%s:        This protocol doesn't support\n"
						 "%s:        multi-port operation!\n",
						 card->devname,ioport,
						 card->devname,card->devname);
				return -EEXIST;
			}
	        }
	}
	/* Make sure I/O port region is available */
	/* (In FreeBSD is done before kernel recompilation) */

	return 0;
}

/*================================================================== 
 * configure_s514_card
 * 
 * For a S514 adapter, check for a possible configuration error in that
 * we are loading an adapter in the same slot as a previously loaded S514
 * card.
 */ 
static int check_s514_conflicts(sdla_t* card,wandev_conf_t* conf, int *irq)
{
	sdla_t*		nxt_card= NULL;
	wan_spinlock_t	s;
	int		i = 0;

	if (card->configured){
		return 0;
	}

	for (i = 0; i < Sangoma_cards_no; i ++){
		nxt_card = &card_array[i];
                if (nxt_card == card){
			continue;
		}

		if (nxt_card->hw == card->hw){
			if ((conf->config_id == WANCONFIG_CHDLC || conf->config_id == WANCONFIG_MPPP) &&
			    (nxt_card->wandev.config_id == WANCONFIG_CHDLC || nxt_card->wandev.config_id == WANCONFIG_MPPP)){ 
			
				if (nxt_card->next != NULL){
					sdla_t*	card1 = (sdla_t*)nxt_card->next;
					/* Trying to run piggibacking with a wrong protocol */
					DEBUG_EVENT("%s: ERROR: Resource busy: CPU %c PCISLOT %d\n",
							card->devname, 
							conf->S514_CPU_no[0],
							conf->PCI_slot_no);
					DEBUG_EVENT("%s:        These resources are already is used\n",
							card->devname);
					DEBUG_EVENT("%s:        by %s and %s devices!\n",
							card->devname, nxt_card->devname, card1->devname);
					return -EBUSY;
				}
				wan_spin_lock_irq(NULL, &s);
				/* ALEX *irq = nxt_card->hw.irq; */
				nxt_card->next		= card;
				card->next		= nxt_card;
				card->wandev.piggyback	= WANOPT_YES;
				wan_spin_lock_init(&card->wandev.lock);
				wan_spin_unlock_irq(NULL, &s);
				break;
			}else{
				/* Trying to run piggibacking with a wrong protocol */
				DEBUG_EVENT("%s: ERROR: Resource busy: CPU %c PCISLOT %d\n",
					card->devname, conf->S514_CPU_no[0],conf->PCI_slot_no);
				DEBUG_EVENT("%s:        This protocol doesn't support\n",
					card->devname);
				DEBUG_EVENT("%s:        multi-port operation!\n",
					card->devname);
				return -EEXIST;
			}
		}
	}

	return 0;
}


/*================================================================== 
 * configure_s514_card
 * 
 * For a S514 adapter, check for a possible configuration error in that
 * we are loading an adapter in the same slot as a previously loaded S514
 * card.
 */ 
static int check_adsl_conflicts(sdla_t* card,wandev_conf_t* conf, int *irq)
{
	sdla_t* 	nxt_card = NULL;
	int		i = 0;

	if (card->configured)
		return 0;

	/* Check for already loaded card with the same IO port, Bus no and IRQ 
	 * If found, copy its hardware configuration and use its
	 * resources (i.e. piggybacking)
	 */
	for (i = 0; i < Sangoma_cards_no; i ++){
		nxt_card = &card_array[i];
	
		if (nxt_card == card)
			continue;

		if (nxt_card->wandev.state == WAN_UNCONFIGURED)
			continue;

		/* Bug Fix:
		 * If we are using auto PCI slot detection:
		 *   assume that new pci slot is the same as 
		 *   the pci slot of the already configured card
		 *   ie. (nxt_card).
		 * 
		 * The reason for this is, when the pci slot is found
		 * the card->hw.PCI_slot_no is updated with the
		 * found slot number 
		 *
		 * Thus, do not use the PCI_slot_no to detect a
		 * piggyback condition if the auto_pci_cfg option
		 * is enabled.
		 */
		
		if (nxt_card->hw == card->hw){
			/* Trying to run piggibacking with a wrong protocol */
			DEBUG_EVENT("%s: ERROR: Resource busy: CPU %c PCISLOT %d\n"
					 "%s:        This protocol doesn't support\n"
					 "%s:        multi-port operation!\n",
						card->devname,
					 	conf->S514_CPU_no[0],
						conf->PCI_slot_no,
						card->devname,
						card->devname);
			return -EEXIST;
		}
	}

	return 0;
}

/*================================================================== 
 * configure_s514_card
 * 
 * For a S514 adapter, check for a possible configuration error in that
 * we are loading an adapter in the same slot as a previously loaded S514
 * card.
 */ 


static int check_aft_conflicts(sdla_t* card,wandev_conf_t* conf, int *irq)
{
	sdla_t*	nxt_card;
	int		i = 0;

	if (wan_test_bit(0,&card->configured))
		return 0;

	
	/* Check for already loaded card with the same IO port, Bus no and IRQ 
	 * If found, copy its hardware configuration and use its
	 * resources (i.e. piggybacking)
	 */
	
	for (i = 0; i < Sangoma_cards_no; i ++){
		nxt_card = &card_array[i];
		if (nxt_card == card)
			continue;

		if (nxt_card->wandev.state == WAN_UNCONFIGURED)
			continue;

		/* Bug Fix:
		 * If we are using auto PCI slot detection:
		 *   assume that new pci slot is the same as 
		 *   the pci slot of the already configured card
		 *   ie. (nxt_card).
		 * 
		 * The reason for this is, when the pci slot is found
		 * the card->hw.PCI_slot_no is updated with the
		 * found slot number 
		 *
		 * Thus, do not use the PCI_slot_no to detect a
		 * piggyback condition if the auto_pci_cfg option
		 * is enabled.
		 */
		if (nxt_card->hw == card->hw){
                	u16 CPU_no;

                	card->hw_iface.getcfg(card->hw, SDLA_CPU, &CPU_no);
			if (CPU_no == conf->S514_CPU_no[0]){
				/* Trying to run piggibacking with a 
				 * wrong protocol */
				DEBUG_EVENT("%s: ERROR: Resource busy: CPU %c PCISLOT %i\n"
					 "%s:        This protocol doesn't support\n"
					 "%s:        multi-port operation!\n",
					 card->devname,
					 conf->S514_CPU_no[0],conf->PCI_slot_no,
					 card->devname,card->devname);
				return -EEXIST;
			}
		}
	}

	return 0;
}


/* 
 * ============================================================================
 * Shut down WAN link driver. 
 * o shut down adapter hardware
 * o release system resources.
 *
 * This function is called by the router when device is being unregistered or
 * when it handles ROUTER_DOWN IOCTL.
 */
static int shutdown (wan_device_t* wandev, wandev_conf_t* conf)
{
	sdla_t* card = NULL;

	WAN_ASSERT(wandev == NULL);
	WAN_ASSERT(wandev->private == NULL);

	card = wandev->private;
	if (wandev->state == WAN_UNCONFIGURED){
		DEBUG_EVENT("%s: Shutdown unconfigured device!\n",
					card->devname);
		return 0;
	}

	DEBUG_CFG("%s: Shutdown device...\n", card->devname);

#if defined(NETGRAPH)
	wan_ng_remove(card);
#endif

	if (wandev->config_id == WANCONFIG_DEBUG){
		card->hw_iface.down(card->hw);
		sdla_unregister(&card->hw, card->devname);
		wandev->state = WAN_UNCONFIGURED;
		return 0;
	}
	/* In case of piggibacking, make sure that 
	 * we never try to shutdown both devices at the same
	 * time, because they depend on one another 
	 */

	/* ALEX VVVV */
#if defined(CONFIG_PRODUCT_WANPIPE_ADSL)
	if (wandev->config_id == WANCONFIG_ADSL && conf != NULL){
		/* Update ADSL VCI/VPI standart list */
		adsl_vcivpi_update(card, conf);
		conf->config_id = wandev->config_id;
	}
#endif
	WAN_DEBUG_END(card);

	if (card->disable_comm){
		card->disable_comm(card);
	}
	/* ALEX ^^^^ */
	wandev->state = WAN_UNCONFIGURED;

	/* Release resources */
	if (card->configured){
	    	release_hw(card);
	}

	active--;		/* decrement number of active cards */

	if (!card->configured){
		sdla_unregister(&card->hw, card->devname);
	}
	DEBUG_CFG("%s: Shutdown finished\n", card->devname);
	return 0;
}

/*
 * ============================================================================
 * release_hw() -
 *
 */
static void release_hw(sdla_t *card)
{
	sdla_t *nxt_card = NULL;

	DEBUG_CFG("%s: Release hw..\n", card->devname);
	/* Check if next device exists */
	if (card->next){
		nxt_card = card->next;
		/* If next device is down release resources...*/
		if (nxt_card->wandev.state != WAN_UNCONFIGURED){
			DEBUG_EVENT("%s: Device %s still running!\n", 
						card->devname, 
						nxt_card->devname);
			if (!card->wandev.piggyback){
		    	        /* Change interrupt vector argument 
				** to piggybacking card 
				*/
				card->hw_iface.restore_intrhand(
						card->hw,
						WAN_FE_LINENO(&card->fe));
				card->hw_iface.set_intrhand(
						nxt_card->hw, 
						sdla_isr, 
						(void*)nxt_card,
						WAN_FE_LINENO(&nxt_card->fe));
				nxt_card->wandev.piggyback = WANOPT_NO;
			}else{
				card->wandev.piggyback = WANOPT_NO;
			}
			/* `Disconnect' two cards: primary & seconadary */
			card->next	 = NULL;
			nxt_card->next	 = NULL;
			card->configured = WANOPT_NO;
		}else{
			DEBUG_EVENT("%s: Should never heppened!\n",
						nxt_card->devname);
		}
	}else{
		DEBUG_EVENT("%s: Master device shutting down!\n", 
					card->devname);
		/* Free irq !!! */
		card->hw_iface.restore_intrhand(
					card->hw,
					WAN_FE_LINENO(&card->fe));
		card->hw_iface.down(card->hw);
		card->configured = 0;
	}
}

/*
 * ============================================================================
 * Driver I/O control. 
 * o verify arguments
 * o perform requested action
 *
 * This function is called when router handles one of the reserved user
 * IOCTLs.  Note that 'arg' stil points to user address space.
 */
static int ioctl (wan_device_t* wandev, u_long cmd, caddr_t arg)
{
	sdla_t* card = NULL;
	int err;

	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -EFAULT;
	}
	if (wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}

	card = wandev->private;

	switch (cmd){ 
	case WANPIPE_DUMP:
		err = ioctl_dump(wandev->private, (void*)arg);
		break;

	case WANPIPE_EXEC:
		err = ioctl_exec(wandev->private, (void*)arg);
		break;

	default:
		err = -EINVAL;
		break;
	}

	return err;
}

/****** Driver IOCTL Handlers ***********************************************/

/* 
 * ============================================================================
 * Dump adapter memory to user buffer.
 * o verify request structure
 * o copy request structure to kernel data space
 * o verify length/offset
 * o verify user buffer
 * o copy adapter memory image to user buffer
 *
 * Note: when dumping memory, this routine switches curent dual-port memory
 *	 vector, so care must be taken to avoid racing conditions.
 */
static int ioctl_dump (sdla_t* card, sdla_dump_t* u_dump)
{
	sdla_dump_t	dump;
	wan_spinlock_t	s;
	char*		data = NULL;
	int 		err = 0;

	if (u_dump == NULL){
		return -EFAULT;
	}
	/* Copy from user space to kernel */
	if(copyin((void*)u_dump, (void*)&dump, sizeof(sdla_dump_t))){
		DEBUG_EVENT("%s: Failed to copy memory from user to kernel space!\n",
						card->devname);
		return -EFAULT;
	}
		
	if (dump.magic != WANPIPE_MAGIC || dump.ptr == NULL){
		return -EFAULT;
	} 

	if ((data = wan_malloc(dump.length)) == NULL){
		DEBUG_EVENT("%s: Failed allocate memory for dump!\n",
						card->devname);
		return -EFAULT;
	}

	wan_spin_lock_irq(NULL, &s);

	card->hw_iface.peek(card->hw, dump.offset, data, dump.length);

	wan_spin_unlock_irq(NULL, &s);
		
	if (copyout((void*)data, (void*)dump.ptr, dump.length)){
		DEBUG_EVENT("%s: Failed to copy memory from kernel to user space!\n",
						card->devname);
		err = -EFAULT; 
	}
	if (data){
		wan_free(data);
	}
	return err;
}

/* 
 * ============================================================================
 * Execute adapter firmware command.
 * o verify request structure
 * o copy request structure to kernel data space
 * o call protocol-specific 'exec' function
 */
static int ioctl_exec (sdla_t* card, sdla_exec_t* u_exec)
{
	sdla_exec_t exec;

	if (card->exec == NULL){
		return -ENODEV;
	}
	
	if (u_exec == NULL){ 
		return -EFAULT;
	}
	if (copyin((void*)u_exec, (void*)&exec, sizeof(sdla_exec_t))){
		return -EFAULT;
	}

	if ((exec.magic != WANPIPE_MAGIC) || (exec.cmd == NULL)){
		return -EINVAL;
	}
	return card->exec(card, exec.cmd, exec.data);
}

/******* Miscellaneous ******************************************************/

/*
 * ============================================================================
 * SDLA Interrupt Service Routine (S514).
 *  o acknoledge SDLA hardware interrupt.
 *  o call protocol-specific interrupt service routine, if any.
 *  o Return 1, for internal error; otherwise - 0;
 */
static void sdla_isr(void *arg)
{
	sdla_t *card = (sdla_t*)arg;

	if (card == NULL){
		DEBUG_ISR("%s: Internal Error (line=%d)\n", 
					__FILE__, __LINE__); 
		return; 
	}

	if (card && card->type == SDLA_AFT){

		if (card->isr){
			card->isr(card);
		}
		return;
	}


	if (card->type == SDLA_ADSL){	/* handler interrupt on S514 */
		if (card->isr){
			card->isr(card);
		}
		return;
	}

	if (card->type == SDLA_S514){	/* handler interrupt on S514 */
       		u32	int_status;
		u16	CPU_no; 
    		int	card_found_for_IRQ;

		card->hw_iface.getcfg(card->hw, SDLA_CPU, &CPU_no);
		if (card->hw_iface.read_int_stat){
			card->hw_iface.read_int_stat(card->hw, &int_status);
		}
		/* check if the interrupt is for this device */
		if (!((unsigned char)int_status & (IRQ_CPU_A | IRQ_CPU_B))){
			return;
		}

		/* If the IRQ is both CPUs on the same addapter, then alter
		 * the interrupt status so as handle one CPU at a time.
		 */
		if (((unsigned char)int_status &(IRQ_CPU_A | IRQ_CPU_B)) ==
					(IRQ_CPU_A | IRQ_CPU_B)){
			int_status &= (CPU_no == SDLA_CPU_A) ? ~IRQ_CPU_B : ~IRQ_CPU_A;
		}
		card_found_for_IRQ = 0;
		/* Check to see that the CPU number for this device corresponds to
		 * the interrupt status read.
		 */
		switch(CPU_no){
		case SDLA_CPU_A:
			if ((unsigned char)int_status & IRQ_CPU_A){
				card_found_for_IRQ = 1;
			}	
			break;
	
		case SDLA_CPU_B:
			if ((unsigned char)int_status & IRQ_CPU_B){
				card_found_for_IRQ = 1;
			}
			break;
		}
		/* exit if the interrupt is for another CPU on the same IRQ */
		if (!card_found_for_IRQ){
			return;
		}
		if (card->wandev.state == WAN_UNCONFIGURED && !card->configured){
			DEBUG_EVENT("%s: IRQ for unconfigured adapter.\n", card->devname);
			if (card->hw_iface.intack){
				card->hw_iface.intack(card->hw, int_status);
			}
			return;
		}
		if (card->in_isr){
			DEBUG_EVENT("%s: interrupt re-entrancy on IRQ %d\n",
						card->devname, 
						card->wandev.irq);
			if (card->hw_iface.intack){
				card->hw_iface.intack(card->hw, int_status);
			}
			return;
		}
		if (card->hw_iface.intack){
			card->hw_iface.intack(card->hw, int_status);
		}

		if (card->isr){
			card->isr(card);
		}
	}else{
		if (card->wandev.state == WAN_UNCONFIGURED){
			return;
		}
		if (card->in_isr){
			DEBUG_EVENT("%s: interrupt re-entrancy on IRQ %d!\n",
					card->devname, card->wandev.irq);
			return;
		}
		if (card->hw_iface.intack){
			card->hw_iface.intack(card->hw, 0x00);
		}
		if (card->isr){
			card->isr(card);
		}
	}
	return;
}

/* 
 * ============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being open.  
 */
void wanpipe_open (sdla_t* card)
{
	++card->open_cnt;
}

/* 
 * ============================================================================
 * This routine is called by the protocol-specific modules when network
 * interface is being closed.  
 */
void wanpipe_close (sdla_t* card)
{
	--card->open_cnt;
}


/*
 ***************************************************************************
 * in_ntoa() - Convert IP address to string  
 */
char *in_ntoa(uint32_t in)
{
	static char buf[18];
	char *p = NULL;

	p = (char*)&in;
	sprintf(buf, "%d.%d.%d.%d",
       		 (p[0] & 0xFF), (p[1] & 0xFF), (p[2] & 0xFF), (p[3] & 0xFF));

	return buf;
}

/*
 **************************************************************************
 * sdla_get() - Copy data from receive buffer to end of mbuf chain allocate 
 * 		additional mbufs as needed. Return pointer to first mbuf in 
 *		chain.
 *
 */
struct mbuf* sdla_get(struct ifnet* ifp, char* data, unsigned totlen)
{
	struct mbuf*	m = NULL;

	m = wan_skb_alloc(totlen);
	m->m_pkthdr.rcvif = ifp;
	wan_skb_copyback(m, wan_skb_len(m), totlen, (caddr_t)data);
	return m;
#if 0
	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL){
		return NULL;
	}
	m->m_pkthdr.rcvif = ifp;
	m->m_pkthdr.len   = totlen;
	len 	      = MHLEN;
	top 	      = NULL;
	mp		      = &top;

	while(totlen > 0){
		if (top) {
			MGET(m, M_DONTWAIT, MT_DATA);
		    	if (m == NULL){
				m_freem(top);
				return NULL;
			}
			len = MHLEN;
		}
		if (top && totlen >= MINCLSIZE){
			MCLGET(m, M_DONTWAIT);
			if (m->m_flags & M_EXT){
				len = MCLBYTES;
			}
		}
		m->m_len = len = min(totlen, len);
		bcopy(data, mtod(m, caddr_t), m->m_len);
		data   += m->m_len;
		totlen -= m->m_len;
		*mp     = m;
		mp	= &m->m_next;
	}
	return top;
#endif
}


static int debugging (wan_device_t* wandev)
{
	sdla_t*			card = (sdla_t*)wandev->private;

	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}
	WAN_DEBUG_START(card);

	return 0;
}

/****** End *****************************************************************/
